"""DICOM-specific functions
inspired by dicom-numpy: https://github.com/innolitics/dicom-numpy
Author: Jacob Reinhold <jcreinhold@gmail.com>
"""

from __future__ import annotations

__all__ = [
    "DICOMDir",
    "DICOMImage",
]

import builtins
import dataclasses
import functools
import io
import logging
import math
import operator
import pathlib
import typing
import warnings
import zipfile

import numpy as np
import numpy.typing as npt

try:
    import pydicom
except ImportError as imp_exn:
    imp_exn_msg = f"pydicom must be installed to use {__name__}."
    raise ImportError(imp_exn_msg) from imp_exn

import pymedio.base as miob
import pymedio.exceptions as mioe
import pymedio.typing as miot
import pymedio.utils as miou

logger = logging.getLogger(__name__)

ORIENTATION_ATOL = 1e-5

T = typing.TypeVar("T")


def _all_float_like(seq: typing.Sequence[builtins.float]) -> builtins.bool:
    return all(isinstance(x, (float, int)) for x in seq)


@dataclasses.dataclass(frozen=True)
class Cosines:
    # dicom-numpy -> dicom_numpy/combine_slices.py
    # ITK -> Modules/IO/GDCM/src/itkGDCMImageIO.cxx
    row: npt.NDArray
    column: npt.NDArray
    slice: npt.NDArray

    def __repr__(self) -> builtins.str:
        return f"Cosines(row={self.row}, column={self.column}, slice={self.slice})"

    @classmethod
    def from_orientation(
        cls: typing.Type[Cosines],
        image_orientation: typing.Sequence[builtins.float] | npt.NDArray,
    ) -> Cosines:
        if isinstance(image_orientation, np.ndarray):
            if image_orientation.size != 6 or image_orientation.ndim != 1:
                raise ValueError("image_orientation must be seq. of 1 dim and len=6.")
        elif len(image_orientation) != 6 or not _all_float_like(image_orientation):
            raise ValueError("image_orientation must be seq. floats with len=6.")
        row_cosine = miou.to_f64(image_orientation[:3])
        column_cosine = miou.to_f64(image_orientation[3:])
        slice_cosine: np.ndarray = np.cross(row_cosine, column_cosine)
        cosines = cls(row_cosine, column_cosine, slice_cosine)
        cosines.writable(False)
        cosines.validate()
        return cosines

    def validate(self) -> None:
        dot_prod = float(np.dot(self.row, self.column).item())
        err_msg_dp = f"Non-orthogonal direction cosines: {self.row}, {self.column}"
        warn_msg_dp = f"Direction cosines aren't quite ortho: {self.row}, {self.column}"
        self._validate_value(dot_prod, err_msg_dp, warn_msg_dp, self._almost_zero)

        row_cosine_norm = float(np.linalg.norm(self.row).item())
        err_msg_rn = f"Row direction cosine's magnitude is not 1: {self.row}"
        warn_msg_rn = f"Row direction cosine's magnitude not quite 1: {self.row}"
        self._validate_value(row_cosine_norm, err_msg_rn, warn_msg_rn, self._almost_one)

        col_cosine_norm = float(np.linalg.norm(self.column).item())
        err_msg_cn = f"Column direction cosine's magnitude is not 1: {self.column}"
        warn_msg_cn = f"Column direction cosine's magnitude not quite 1: {self.column}"
        self._validate_value(col_cosine_norm, err_msg_cn, warn_msg_cn, self._almost_one)

    def writable(self, value: builtins.bool, /) -> None:
        self.row.flags.writeable = value
        self.column.flags.writeable = value
        self.slice.flags.writeable = value

    @staticmethod
    def _validate_value(
        value: builtins.float,
        err_msg: builtins.str,
        warn_msg: builtins.str,
        check_func: typing.Callable[..., builtins.bool],
    ) -> None:
        if not check_func(value, atol=1e-4):
            raise mioe.DicomImportException(value, err_msg)
        elif not check_func(value, atol=1e-8):
            warnings.warn(warn_msg)

    @staticmethod
    def _almost_zero(value: builtins.float, *, atol: builtins.float) -> builtins.bool:
        return math.isclose(value, 0.0, abs_tol=atol)

    @staticmethod
    def _almost_one(value: builtins.float, *, atol: builtins.float) -> builtins.bool:
        return math.isclose(value, 1.0, abs_tol=atol)


@dataclasses.dataclass(frozen=True)
class SortedSlices:
    slices: typing.Tuple[pydicom.Dataset, ...]
    indices: typing.Tuple[builtins.int, ...]  # mapping to get original slice order
    positions: typing.Tuple[builtins.float, ...]
    cosines: Cosines

    def __repr__(self) -> builtins.str:
        return f"SortedSlices(n_slices={len(self)}, cosines={self.cosines!r})"

    def __len__(self) -> builtins.int:
        _len = len(self.slices)
        exn_msg: typing.List[builtins.str] = []
        if _len != (ind_len := len(self.indices)):
            exn_msg.append(f"num slices {_len} != num indices {ind_len}")
        if _len != (pos_len := len(self.positions)):
            exn_msg.append(f"num slices {_len} != num positions {pos_len}")
        if exn_msg:
            raise RuntimeError(" ".join(exn_msg))
        return _len

    @classmethod
    def from_datasets(
        cls, slice_datasets: typing.Sequence[pydicom.Dataset]
    ) -> SortedSlices:
        """sort list of pydicom datasets into the correct order"""
        if not slice_datasets:
            raise ValueError("slice_datasets empty")
        image_orientation = miou.to_f64(slice_datasets[0].ImageOrientationPatient)
        cosines = Cosines.from_orientation(image_orientation)
        cs = cosines.slice
        imps = (miou.to_f64(sd.ImagePositionPatient) for sd in slice_datasets)
        positions = (np.dot(cs, imp).item() for imp in imps)
        _sorted = typing.cast(
            typing.Iterable[typing.Tuple[builtins.int, builtins.float]],
            sorted(enumerate(positions), key=operator.itemgetter(1)),
        )
        sorted_indices, sorted_positions = miou.unzip(_sorted)
        sorted_slice_datasets = tuple(slice_datasets[i] for i in sorted_indices)
        return cls(
            slices=sorted_slice_datasets,
            indices=sorted_indices,
            positions=sorted_positions,
            cosines=cosines,
        )

    def check_nonuniformity(
        self,
        *,
        max_nonuniformity: builtins.float = 5e-4,
        fail_outside_max_nonuniformity: builtins.bool = True,
        missing_slices_cutoff: builtins.float = 1e-1,
    ) -> None:
        if len(self) > 1:
            diffs = np.diff(self.positions)
            if not np.allclose(diffs, diffs[0], atol=0.0, rtol=max_nonuniformity):
                msg = f"The slice spacing is non-uniform. Slice spacings:\n{diffs}"
                if fail_outside_max_nonuniformity:
                    raise mioe.OutsideMaxNonUniformity(msg)
                else:
                    warnings.warn(msg)
            if not np.allclose(diffs, diffs[0], atol=0.0, rtol=missing_slices_cutoff):
                msg = "There appear to be missing slices."
                raise mioe.MissingSlicesException(msg)

    def remove_anomalous_slices(
        self, *, strict_unique: builtins.bool = True
    ) -> SortedSlices:
        to_float_tuple = lambda xs: tuple(float(x) for x in xs)  # noqa: E731
        orientations = [to_float_tuple(s.ImageOrientationPatient) for s in self.slices]
        if strict_unique:
            unq_oris: np.ndarray
            unq_oris, counts = np.unique(orientations, axis=0, return_counts=True)
            most_common_orientation = unq_oris[np.argmax(counts)]
        else:
            approx_unique_orientations = self._approx_unique(orientations)
            most_common_orientation = approx_unique_orientations[-1]
        out = [
            (s, idx, pos)
            for s, idx, pos, o in self._zip(orientations)
            if np.allclose(o, most_common_orientation, atol=ORIENTATION_ATOL)
        ]
        new_slices, new_indices, new_positions = miou.unzip(out)
        if (n_removed := (len(self) - len(new_slices))) > 1:
            warnings.warn(f"{n_removed} anomalous images removed.")
        elif n_removed < 0:
            raise RuntimeError("Images added in remove image func. Report error.")
        new_image_orientation = miou.to_f64(new_slices[0].ImageOrientationPatient)
        new_cosines = Cosines.from_orientation(new_image_orientation)
        return SortedSlices(
            slices=new_slices,
            indices=new_indices,
            positions=new_positions,
            cosines=new_cosines,
        )

    @functools.cached_property
    def patient_position(self) -> npt.NDArray:
        return miou.to_f64(self.slices[0].ImagePositionPatient)

    @functools.cached_property
    def slice_spacing(self) -> builtins.float:
        spacing: builtins.float
        if len(self) > 1:
            slice_positions_diffs = np.diff(np.sort(self.positions))
            # avg. b/c that's what ITK seems to use, so use for consistency
            spacing = float(np.mean(slice_positions_diffs).item())
        elif len(self) == 1:
            spacing = float(getattr(self.slices[0], "SpacingBetweenSlices", 0))
        else:
            raise RuntimeError("slice_datasets must contain at least one dicom image")
        return spacing

    @functools.cached_property
    def affine(self) -> npt.NDArray:
        row_spacing, column_spacing = self.slices[0].PixelSpacing
        transform: np.ndarray = np.identity(4, dtype=np.float64)
        slice_spacing = self.slice_spacing or 1.0
        transform[:3, 0] = self.cosines.row * column_spacing
        transform[:3, 1] = self.cosines.column * row_spacing
        transform[:3, 2] = self.cosines.slice * slice_spacing
        transform[:3, 3] = self.patient_position
        transform_ras: npt.NDArray = np.dot(miou.flipxy_44(), transform)
        return transform_ras

    @staticmethod
    def _approx_unique(
        values: typing.Sequence[T],
        *,
        atol: builtins.float = ORIENTATION_ATOL,
    ) -> typing.Tuple[T, ...]:
        # TODO: improve computational efficiency
        # TODO: fix bad init -> bad result
        if not values:
            return tuple()
        approx_unique: typing.Dict[T, builtins.int] = dict()
        for val in values:
            min_dist = np.inf
            min_dist_val = None
            for target_val in approx_unique.keys():
                np_val = miou.to_f64(typing.cast(npt.ArrayLike, val))
                np_tgt_val = miou.to_f64(typing.cast(npt.ArrayLike, target_val))
                if np.allclose(np_val, np_tgt_val, atol=atol):
                    dist = np.linalg.norm(np_val - np_tgt_val).item()
                    if dist < min_dist:
                        min_dist = dist
                        min_dist_val = target_val
            if min_dist_val is None:
                approx_unique[val] = 1
            else:
                approx_unique[min_dist_val] += 1
        approx_unq_arrs, _ = miou.unzip(
            sorted(
                ((arr, count) for arr, count in approx_unique.items()),
                key=operator.itemgetter(1),
            )
        )
        return approx_unq_arrs

    def _zip(
        self, *args: typing.Iterable[typing.Any]
    ) -> typing.Iterable[typing.Tuple[typing.Any, ...]]:
        return zip(self.slices, self.indices, self.positions, *args)


@dataclasses.dataclass(frozen=True)
class DICOMDir:
    slices: typing.Tuple[pydicom.Dataset, ...]
    positions: typing.Tuple[builtins.float, ...]
    slice_spacing: builtins.float
    affine: npt.NDArray
    paths: typing.Tuple[miot.PathLike, ...] | None = None

    def __len__(self) -> builtins.int:
        _len = len(self.slices)
        if self.paths is not None and _len != (path_len := len(self.paths)):
            raise RuntimeError(f"num slices {_len} != num paths {path_len}")
        return _len

    @classmethod
    def from_datasets(
        cls: typing.Type[DICOMDir],
        datasets: typing.Sequence[pydicom.Dataset],
        *,
        paths: typing.Sequence[miot.PathLike] | None = None,
        max_nonuniformity: builtins.float = 5e-4,
        fail_outside_max_nonuniformity: builtins.bool = True,
        remove_anomalous_images: builtins.bool = True,
    ) -> DICOMDir:
        if not datasets:
            msg = "Must provide at least one image DICOM dataset"
            raise mioe.DicomImportException(msg)
        sorted_slices = SortedSlices.from_datasets(datasets)
        sorted_slices.check_nonuniformity(
            max_nonuniformity=max_nonuniformity,
            fail_outside_max_nonuniformity=fail_outside_max_nonuniformity,
        )
        if remove_anomalous_images:
            sorted_slices = sorted_slices.remove_anomalous_slices()
        positions = tuple(sorted(sorted_slices.positions))
        idxs = sorted_slices.indices
        dicom_dir = cls(
            slices=sorted_slices.slices,
            positions=positions,
            slice_spacing=sorted_slices.slice_spacing,
            affine=sorted_slices.affine,
            paths=None if paths is None else tuple(paths[i] for i in idxs),
        )
        dicom_dir.writable(False)
        dicom_dir.validate()
        return dicom_dir

    @classmethod
    def from_path(
        cls: typing.Type[DICOMDir],
        dicom_path: miot.PathLike | typing.Iterable[miot.PathLike],
        *,
        max_nonuniformity: builtins.float = 5e-4,
        fail_outside_max_nonuniformity: builtins.bool = True,
        remove_anomalous_images: builtins.bool = True,
        defer_size: builtins.str | builtins.int | None = "1 KB",
        extension: builtins.str = ".dcm",
    ) -> DICOMDir:
        paths: typing.Tuple[miot.PathLike, ...]
        if (
            isinstance(dicom_path, (builtins.str, pathlib.Path))
            and (_dcm_dir := pathlib.Path(dicom_path)).is_dir()
        ):
            paths = tuple(sorted(_dcm_dir.glob(f"*{extension}")))
        elif (
            not isinstance(dicom_path, (builtins.str, pathlib.Path))
            and miou.is_iterable(dicom_path)
            and all(str(p).endswith(f".{extension}") for p in dicom_path)  # type: ignore[union-attr]  # noqa: E501
        ):
            paths = tuple(dicom_path)  # type: ignore[arg-type]
        else:
            raise ValueError("dicom_dir must be path to a dir. or a list of dcm paths")
        images = tuple(pydicom.dcmread(path, defer_size=defer_size) for path in paths)
        return cls.from_datasets(
            typing.cast(typing.List[pydicom.Dataset], images),
            paths=paths,
            max_nonuniformity=max_nonuniformity,
            fail_outside_max_nonuniformity=fail_outside_max_nonuniformity,
            remove_anomalous_images=remove_anomalous_images,
        )

    @classmethod
    def from_zipped_stream(
        cls: typing.Type[DICOMDir],
        data_stream: typing.BinaryIO,
        *,
        max_nonuniformity: builtins.float = 5e-4,
        fail_outside_max_nonuniformity: builtins.bool = True,
        remove_anomalous_images: builtins.bool = True,
        encryption_key: builtins.bytes | builtins.str | None = None,
        **zip_kwargs: typing.Any,
    ) -> DICOMDir:
        if encryption_key is not None:
            try:
                import cryptography.fernet as crypto
            except (ModuleNotFoundError, ImportError) as crypto_imp_exn:
                msg = "If encryption key provided, cryptography package required."
                raise RuntimeError(msg) from crypto_imp_exn
            fernet = crypto.Fernet(encryption_key)
            data_stream.seek(0)
            data_stream = io.BytesIO(fernet.decrypt(data_stream.read()))
        with zipfile.ZipFile(data_stream, mode="r", **zip_kwargs) as zf:
            datasets = cls.dicom_datasets_from_zip(zf)
        return cls.from_datasets(
            datasets,
            max_nonuniformity=max_nonuniformity,
            fail_outside_max_nonuniformity=fail_outside_max_nonuniformity,
            remove_anomalous_images=remove_anomalous_images,
        )

    @staticmethod
    def dicom_datasets_from_zip(
        zip_file: zipfile.ZipFile,
    ) -> typing.List[pydicom.Dataset]:
        datasets: typing.List[pydicom.Dataset] = []
        for name in zip_file.namelist():
            if name.endswith("/"):
                continue  # skip directories
            with zip_file.open(name, mode="r") as f:
                try:
                    datasets.append(pydicom.dcmread(f))  # type: ignore[arg-type]
                except pydicom.errors.InvalidDicomError as e:
                    msg = f"Skipping invalid DICOM file '{name}': {e}"
                    logger.info(msg)
        if not datasets:
            msg = "Zipfile does not contain any valid DICOM files"
            raise mioe.DicomImportException(msg)
        return datasets

    def validate(self) -> None:
        invariant_properties = frozenset(
            (
                "BitsAllocated",
                "Columns",
                "Modality",
                "PixelRepresentation",
                "PixelSpacing",
                "Rows",
                "SamplesPerPixel",
                "SeriesInstanceUID",
                "SOPClassUID",
            )
        )
        for property_name in invariant_properties:
            self._slice_attribute_equal(property_name)
        self._slice_attribute_almost_equal(
            "ImageOrientationPatient", atol=ORIENTATION_ATOL
        )

    def writable(self, value: builtins.bool, /) -> None:
        self.affine.flags.writeable = value

    def _slice_attribute_equal(self, property_name: builtins.str) -> None:
        initial_value = getattr(self.slices[0], property_name, None)
        for dataset in self.slices[1:]:
            value = getattr(dataset, property_name, None)
            if value != initial_value:
                msg = "All slices must have the same value for "
                msg += f"'{property_name}': {value} != {initial_value}"
                raise mioe.DicomImportException(msg)

    def _slice_attribute_almost_equal(
        self,
        property_name: builtins.str,
        *,
        atol: builtins.float,
    ) -> None:
        initial_value: miot.SupportsArray | None
        initial_value = getattr(self.slices[0], property_name, None)
        for dataset in self.slices[1:]:
            value: miot.SupportsArray | None
            value = getattr(dataset, property_name, None)
            if value is None or initial_value is None:
                msg = f"All slices must contain the attribute {property_name}"
                raise mioe.DicomImportException(msg)
            if not np.allclose(value, initial_value, atol=atol):
                msg = "All slices must have the same value for "
                msg += f"'{property_name}' within '{atol}': {value} != {initial_value}"
                raise mioe.DicomImportException(msg)

    @staticmethod
    def _is_dicomdir(dataset: pydicom.Dataset) -> builtins.bool:
        media_sop_class: builtins.str | None
        media_sop_class = getattr(dataset, "MediaStorageSOPClassUID", None)
        result: builtins.bool = media_sop_class == "1.2.840.10008.1.3.10"
        return result


class DICOMImage(miob.BasicImage):
    @classmethod
    def from_dicomdir(
        cls: typing.Type[DICOMImage],
        dicom_dir: DICOMDir,
        *,
        rescale: builtins.bool | None = None,
        rescale_dtype: npt.DTypeLike = np.float32,
        order: typing.Literal["F", "C"] | None = None,
    ) -> DICOMImage:
        data = cls._merge_slice_pixel_arrays(
            dicom_dir.slices, rescale=rescale, rescale_dtype=rescale_dtype, order=order
        )
        return cls(data=data, affine=dicom_dir.affine)

    @classmethod
    def from_path(
        cls: typing.Type[DICOMImage],
        dicom_path: miot.PathLike | typing.Iterable[miot.PathLike],
        *,
        max_nonuniformity: builtins.float = 5e-4,
        fail_outside_max_nonuniformity: builtins.bool = True,
        remove_anomalous_images: builtins.bool = True,
        rescale: builtins.bool | None = None,
        rescale_dtype: npt.DTypeLike = np.float32,
        order: typing.Literal["F", "C"] | None = None,
        extension: builtins.str = ".dcm",
    ) -> DICOMImage:
        dicomdir = DICOMDir.from_path(
            dicom_path,
            max_nonuniformity=max_nonuniformity,
            fail_outside_max_nonuniformity=fail_outside_max_nonuniformity,
            remove_anomalous_images=remove_anomalous_images,
            extension=extension,
        )
        return cls.from_dicomdir(
            dicomdir, rescale=rescale, rescale_dtype=rescale_dtype, order=order
        )

    @classmethod
    def from_zipped_stream(
        cls: typing.Type[DICOMImage],
        data_stream: typing.BinaryIO,
        *,
        max_nonuniformity: builtins.float = 5e-4,
        fail_outside_max_nonuniformity: builtins.bool = True,
        remove_anomalous_images: builtins.bool = True,
        encryption_key: builtins.bytes | builtins.str | None = None,
        rescale: builtins.bool | None = None,
        rescale_dtype: npt.DTypeLike = np.float32,
        order: typing.Literal["F", "C"] | None = None,
        **zip_kwargs: typing.Any,
    ) -> DICOMImage:
        dicomdir = DICOMDir.from_zipped_stream(
            data_stream,
            max_nonuniformity=max_nonuniformity,
            fail_outside_max_nonuniformity=fail_outside_max_nonuniformity,
            remove_anomalous_images=remove_anomalous_images,
            encryption_key=encryption_key,
            **zip_kwargs,
        )
        return cls.from_dicomdir(
            dicomdir, rescale=rescale, rescale_dtype=rescale_dtype, order=order
        )

    @classmethod
    def _merge_slice_pixel_arrays(
        cls: typing.Type[DICOMImage],
        slices: typing.Sequence[pydicom.Dataset],
        *,
        rescale: builtins.bool | None = None,
        rescale_dtype: npt.DTypeLike = np.float32,
        order: typing.Literal["F", "C"] | None = None,
    ) -> npt.NDArray:
        if rescale is None:
            rescale = any(cls._requires_rescaling(d) for d in slices)

        first_dataset = slices[0]
        slice_dtype = first_dataset.pixel_array.dtype
        slice_shape = first_dataset.pixel_array.T.shape
        slice_order = "F" if first_dataset.pixel_array.T.flags.f_contiguous else "C"
        num_slices = len(slices)

        voxels_shape = slice_shape + (num_slices,)
        voxels_dtype = rescale_dtype if rescale else slice_dtype
        voxels = np.empty(
            voxels_shape,
            dtype=voxels_dtype,
            order=typing.cast(typing.Literal["F", "C"], slice_order),
        )

        for k, dataset in enumerate(slices):
            pixel_array = dataset.pixel_array.T.astype(voxels_dtype, copy=False)
            if rescale:
                slope = float(getattr(dataset, "RescaleSlope", 1.0))
                intercept = float(getattr(dataset, "RescaleIntercept", 0.0))
                if slope != 1.0:
                    pixel_array *= slope
                if intercept != 0.0:
                    pixel_array += intercept
            voxels[..., k] = pixel_array

        if order is not None:
            if order == "C":
                voxels = np.ascontiguousarray(voxels)
            elif order == "F":
                voxels = np.asfortranarray(voxels)
            else:
                msg = f"If order given, must be either 'F' or 'C'. Got {order}."
                raise ValueError(msg)

        if voxels.dtype != voxels_dtype:
            exn_msg = f"voxels.dtype {voxels.dtype} != requested dtype {voxels_dtype}"
            raise RuntimeError(exn_msg)
        return voxels

    @staticmethod
    def _requires_rescaling(dataset: pydicom.Dataset) -> builtins.bool:
        return hasattr(dataset, "RescaleSlope") or hasattr(dataset, "RescaleIntercept")



"""Image class to hold a medical image
Taken from torchio and modified to use eager
load and use np.ndarray instead of torch.tensor
https://github.com/fepegar/torchio
Author: Jacob Reinhold <jcreinhold@gmail.com>
"""

from __future__ import annotations

__all__ = ["Image"]

import builtins
import typing
import zipfile

import numpy as np
import numpy.typing as npt

try:
    import nibabel as nib
    import SimpleITK as sitk
except ImportError as imp_exn:
    imp_exn_msg = f"NiBabel and SimpleITK must be installed to use {__name__}."
    raise ImportError(imp_exn_msg) from imp_exn

import pymedio.base as miob
import pymedio.dicom as miod
import pymedio.functional as miof
import pymedio.typing as miot


class Image(miob.BasicImage):
    @property
    def str_properties(self) -> typing.List[builtins.str]:
        props = super().str_properties
        props += [f"orientation: {''.join(self.orientation)}+"]
        return props

    @property
    def orientation(self) -> typing.Tuple[builtins.str, builtins.str, builtins.str]:
        """Orientation codes."""
        codes: typing.Tuple[builtins.str, builtins.str, builtins.str]
        codes = nib.aff2axcodes(self.affine)
        return codes

    @property
    def bounds(self) -> npt.NDArray:
        """Position of centers of voxels in smallest and largest indices."""
        ini = 0, 0, 0
        fin: np.ndarray = np.asarray(self.shape) - 1
        point_ini = nib.affines.apply_affine(self.affine, ini)
        point_fin = nib.affines.apply_affine(self.affine, fin)
        return np.asarray((point_ini, point_fin))

    def axis_name_to_index(self, axis: builtins.str) -> builtins.int:
        # Top and bottom are used for the vertical 2D axis as the use of
        # Height vs Horizontal might be ambiguous
        if not isinstance(axis, str):
            raise ValueError("Axis must be a string")
        axis = axis[0].upper()

        if axis in "TB":  # Top, Bottom
            return -2
        else:
            try:
                index = self.orientation.index(axis)
            except ValueError:
                index = self.orientation.index(self.flip_axis(axis))
            # Return negative indices so that it does not matter whether we
            # refer to spatial dimensions or not
            index = -3 + index
            return index

    @staticmethod
    def flip_axis(axis: builtins.str) -> builtins.str:
        labels = "LRPAISTBDV"
        first = labels[::2]
        last = labels[1::2]
        flip_dict = {a: b for a, b in zip(first + last, last + first)}
        axis = axis[0].upper()
        flipped_axis = flip_dict.get(axis)
        if flipped_axis is None:
            values = ", ".join(labels)
            message = f"Axis not understood. Please use one of: {values}"
            raise ValueError(message)
        return flipped_axis

    def get_bounds(self) -> miot.Bounds:
        """Get minimum and maximum world coordinates occupied by the image."""
        first_index = 3 * (-0.5,)
        last_index: np.ndarray = np.asarray(self.shape) - 0.5
        first_point = nib.affines.apply_affine(self.affine, first_index)
        last_point = nib.affines.apply_affine(self.affine, last_index)
        array: np.ndarray = np.asarray((first_point, last_point))
        bounds_x, bounds_y, bounds_z = array.T.tolist()
        return bounds_x, bounds_y, bounds_z

    def save(self, path: miot.PathLike, *, squeeze: builtins.bool = True) -> None:
        miof.write_image(
            self.view(np.ndarray),
            self.affine,
            path,
            squeeze=squeeze,
        )

    def to_filename(self, path: miot.PathLike) -> None:
        self.save(path, squeeze=False)

    def get_center(self, lps: builtins.bool = False) -> miot.TripletFloat:
        """Get image center in RAS+ or LPS+ coordinates"""
        size: np.ndarray = np.asarray(self.shape)
        center_index = (size - 1) / 2
        r, a, s = nib.affines.apply_affine(self.affine, center_index)
        return (-r, -a, s) if lps else (r, a, s)

    @classmethod
    def from_path(
        cls: typing.Type[Image],
        path: miot.PathLike,
        *,
        dtype: npt.DTypeLike = np.float32,
        mmap: builtins.bool = False,
    ) -> Image:
        data, affine = miof.read_image(path, dtype=dtype, mmap=mmap)
        return cls(data=data, affine=affine)

    @classmethod
    def from_stream(
        cls: typing.Type[Image],
        data_stream: typing.BinaryIO,
        *,
        dtype: npt.DTypeLike = np.float32,
        gzipped: builtins.bool = False,
        image_class: miof.NibabelImageClass | None = None,
    ) -> Image:
        data, affine = miof.read_image_from_stream(
            data_stream, dtype=dtype, gzipped=gzipped, image_class=image_class
        )
        return cls(data=data, affine=affine)

    @classmethod
    def from_zipped_stream(
        cls: typing.Type[Image],
        data_stream: typing.BinaryIO,
        *,
        dtype: npt.DTypeLike = np.float32,
        gzipped: builtins.bool = False,
        image_class: miof.NibabelImageClass | None = None,
        **zip_kwargs: typing.Any,
    ) -> Image:
        with zipfile.ZipFile(data_stream, "r", **zip_kwargs) as zf:
            names = [name for name in zf.namelist() if not name.endswith("/")]
            if (n := len(names)) != 1:
                msg = f"{n} files in zipped archive. This constructor requires only 1."
                raise RuntimeError(msg)
            name = names[0]
            with zf.open(name, mode="r") as f:
                return cls.from_stream(
                    typing.cast(typing.BinaryIO, f),
                    dtype=dtype,
                    gzipped=gzipped,
                    image_class=image_class,
                )

    @classmethod
    def from_sitk(cls: typing.Type[Image], sitk_image: sitk.Image) -> Image:
        data, affine = miof.sitk_to_array(sitk_image)
        return cls(data=data, affine=affine)

    @classmethod
    def from_dicom_image(
        cls: typing.Type[Image], dicom_image: miod.DICOMImage
    ) -> Image:
        return cls(data=dicom_image, affine=dicom_image.affine)

    @classmethod
    def from_dicom_zipped_stream(
        cls: typing.Type[Image],
        data_stream: typing.BinaryIO,
        *,
        max_nonuniformity: builtins.float = 5e-4,
        fail_outside_max_nonuniformity: builtins.bool = True,
        remove_anomalous_images: builtins.bool = True,
        encryption_key: builtins.bytes | builtins.str | None = None,
        rescale: builtins.bool | None = None,
        rescale_dtype: npt.DTypeLike = np.float32,
    ) -> Image:
        dicom_image = miod.DICOMImage.from_zipped_stream(
            data_stream,
            max_nonuniformity=max_nonuniformity,
            fail_outside_max_nonuniformity=fail_outside_max_nonuniformity,
            remove_anomalous_images=remove_anomalous_images,
            encryption_key=encryption_key,
            rescale=rescale,
            rescale_dtype=rescale_dtype,
        )
        return cls.from_dicom_image(dicom_image)

    def to_sitk(self, **kwargs: builtins.bool) -> sitk.Image:
        """Get the image as an instance of :class:`sitk.Image`."""
        return miof.array_to_sitk(self.view(np.ndarray), self.affine, **kwargs)

    def to_nibabel(self) -> nib.Nifti1Image:
        return nib.Nifti1Image(self.view(np.ndarray), self.affine)



"""Project exceptions
Author: Jacob Reinhold <jcreinhold@gmail.com>
"""


class MedIOException(Exception):
    ...


class DicomImportException(MedIOException):
    ...


class OutsideMaxNonUniformity(MedIOException):
    ...


class MissingSlicesException(MedIOException):
    ...



"""Tests for `medio` package."""

from __future__ import annotations

import builtins
import gzip
import io
import pathlib
import zipfile

import cryptography.fernet as crypto
import nibabel as nib
import numpy as np
import pydicom
import pytest
import SimpleITK as sitk
from pydicom.data import get_testdata_file

import pymedio.dicom as miod
import pymedio.image as mioi

TEST_IMAGE_NAME = "CT_small.dcm"
TEST_IMAGE_SHAPE = (128, 128)
NUM_DUPLICATES = 5
NIFTI_IMAGE_SHAPE = (2, 2, 2)
DIST = 1e-05


@pytest.fixture(scope="session")
def dicom_image_path() -> pathlib.Path:
    path = get_testdata_file(TEST_IMAGE_NAME, read=False)
    assert isinstance(path, builtins.str)
    return pathlib.Path(path)


@pytest.fixture(scope="session")
def dicom_image_dir(
    tmp_path_factory: pytest.TempPathFactory, dicom_image_path: pathlib.Path
) -> pathlib.Path:
    dcm_dir = tmp_path_factory.mktemp("dicom").resolve(strict=True)
    path = dicom_image_path.resolve(strict=True)
    dcm = pydicom.dcmread(str(path))
    position = dcm.ImagePositionPatient
    for i in range(NUM_DUPLICATES):
        position[2] += DIST
        dcm.ImagePositionPatient = position
        pydicom.dcmwrite(dcm_dir / f"{i}.dcm", dcm)
    return dcm_dir


@pytest.fixture(scope="session")
def dicom_image(dicom_image_path: pathlib.Path) -> pydicom.Dataset:
    return pydicom.dcmread(dicom_image_path)


@pytest.fixture(scope="session")
def nifti() -> nib.Nifti1Image:
    data: np.ndarray = np.random.randn(*NIFTI_IMAGE_SHAPE).astype(np.float32)
    return nib.Nifti1Image(data, np.eye(4))


def _make_nifti_path(
    tmp_path_factory: pytest.TempPathFactory,
    nifti: nib.Nifti1Image,
    extension: builtins.str,
) -> pathlib.Path:
    dcm_dir = tmp_path_factory.mktemp(extension.replace(".", "")).resolve(strict=True)
    path = dcm_dir / ("test" + extension)
    nifti.to_filename(path)
    return path


@pytest.fixture(scope="session")
def nifti_gzipped_image_path(
    tmp_path_factory: pytest.TempPathFactory, nifti: nib.Nifti1Image
) -> pathlib.Path:
    return _make_nifti_path(tmp_path_factory, nifti, ".nii.gz")


@pytest.fixture(scope="session")
def nifti_image_path(
    tmp_path_factory: pytest.TempPathFactory, nifti: nib.Nifti1Image
) -> pathlib.Path:
    return _make_nifti_path(tmp_path_factory, nifti, ".nii")


@pytest.fixture(scope="session")
def zipped_dicom_path(
    tmp_path_factory: pytest.TempPathFactory, dicom_image: pydicom.Dataset
) -> pathlib.Path:
    zipped_path = tmp_path_factory.getbasetemp().resolve(strict=True) / "dcm.zip"
    with io.BytesIO() as buffer:
        pydicom.dcmwrite(buffer, dicom_image)
        buffer.seek(0)
        with zipfile.ZipFile(zipped_path, mode="w") as zf:
            zf.writestr("test", buffer.getvalue())
    return zipped_path


@pytest.fixture(scope="session")
def encryption_key() -> builtins.bytes:
    return crypto.Fernet.generate_key()


@pytest.fixture(scope="session")
def zipped_encrypted_dicom_path(
    tmp_path_factory: pytest.TempPathFactory,
    dicom_image: pydicom.Dataset,
    encryption_key: builtins.bytes,
) -> pathlib.Path:
    zipped_path = tmp_path_factory.getbasetemp().resolve(strict=True) / "dcm.zip"
    fernet = crypto.Fernet(encryption_key)
    with io.BytesIO() as zip_buffer:
        with zipfile.ZipFile(zip_buffer, mode="w") as zf:
            with io.BytesIO() as image_buffer:
                pydicom.dcmwrite(image_buffer, dicom_image)
                zf.writestr("test", image_buffer.getvalue())
        encrypted = fernet.encrypt(zip_buffer.getvalue())
    with open(zipped_path, "wb") as f:
        f.write(encrypted)
    return zipped_path


@pytest.fixture(scope="session")
def zipped_nifti_path(
    tmp_path_factory: pytest.TempPathFactory, nifti: nib.Nifti1Image
) -> pathlib.Path:
    zipped_path = tmp_path_factory.getbasetemp().resolve(strict=True) / "nii.zip"
    with io.BytesIO() as buffer:
        file_map = nib.Nifti1Image.make_file_map({"image": buffer, "header": buffer})
        nifti.to_file_map(file_map)
        with zipfile.ZipFile(zipped_path, mode="w") as zf:
            zf.writestr("test", buffer.getvalue())
    return zipped_path


@pytest.fixture(scope="session")
def zipped_nifti_gzipped_path(
    tmp_path_factory: pytest.TempPathFactory, nifti: nib.Nifti1Image
) -> pathlib.Path:
    zipped_path = tmp_path_factory.getbasetemp().resolve(strict=True) / "niigz.zip"
    with io.BytesIO() as buffer:
        file_map = nib.Nifti1Image.make_file_map({"image": buffer, "header": buffer})
        nifti.to_file_map(file_map)
        gz = gzip.compress(buffer.getvalue())
        with zipfile.ZipFile(zipped_path, mode="w") as zf:
            zf.writestr("test", gz)
    return zipped_path


@pytest.fixture
def image() -> mioi.Image:
    return mioi.Image(np.zeros(NIFTI_IMAGE_SHAPE, dtype=np.float32))


def test_dicomdir_from_path(dicom_image_dir: pathlib.Path) -> None:
    dcmdir = miod.DICOMDir.from_path(dicom_image_dir)
    dcmdir.validate()
    assert len(dcmdir) == NUM_DUPLICATES


def test_dicomdir_from_dataset(dicom_image: pydicom.Dataset) -> None:
    datasets = [dicom_image]
    dcmdir = miod.DICOMDir.from_datasets(datasets)
    dcmdir.validate()
    assert len(dcmdir) == 1


def test_dicomdir_from_zipped_stream(zipped_dicom_path: pathlib.Path) -> None:
    with open(zipped_dicom_path, "rb") as f:
        dcmdir = miod.DICOMDir.from_zipped_stream(f)
    dcmdir.validate()
    assert len(dcmdir) == 1


def test_dicomimage_from_path(dicom_image_dir: pathlib.Path) -> None:
    image = miod.DICOMImage.from_path(dicom_image_dir)
    assert image.shape == (TEST_IMAGE_SHAPE + (NUM_DUPLICATES,))


def test_affine_in_image_vs_dicomimage(dicom_image_dir: pathlib.Path) -> None:
    dcm_image = miod.DICOMImage.from_path(dicom_image_dir)
    image = mioi.Image.from_path(dicom_image_dir)
    assert dcm_image.shape == image.shape
    assert np.allclose(
        dcm_image.affine, image.affine
    ), f"\nDICOMImage:\n{dcm_image.affine}\nImage:\n{image.affine}"


def test_affine_vs_tio(dicom_image_dir: pathlib.Path) -> None:
    tio = pytest.importorskip("torchio")
    dcm_image = miod.DICOMImage.from_path(dicom_image_dir)
    image = mioi.Image.from_path(dicom_image_dir)
    tio_image = tio.ScalarImage(dicom_image_dir)
    assert np.allclose(image.affine, tio_image.affine)
    assert np.allclose(dcm_image.affine, tio_image.affine)


def test_dicomimage_from_zipped_stream(zipped_dicom_path: pathlib.Path) -> None:
    with open(zipped_dicom_path, "rb") as f:
        image = miod.DICOMImage.from_zipped_stream(f)
    assert image.shape == (TEST_IMAGE_SHAPE + (1,))


def test_dicom_image_from_zipped_stream(zipped_dicom_path: pathlib.Path) -> None:
    with open(zipped_dicom_path, "rb") as f:
        image = mioi.Image.from_dicom_zipped_stream(f)
    assert image.shape == (TEST_IMAGE_SHAPE + (1,))


def test_dicom_image_from_zipped_stream_encrypted(
    zipped_encrypted_dicom_path: pathlib.Path, encryption_key: builtins.bytes
) -> None:
    with open(zipped_encrypted_dicom_path, "rb") as f:
        image = mioi.Image.from_dicom_zipped_stream(f, encryption_key=encryption_key)
    assert image.shape == (TEST_IMAGE_SHAPE + (1,))


def test_nifti_image_from_path(nifti_image_path: pathlib.Path) -> None:
    image = mioi.Image.from_path(nifti_image_path)
    assert image.shape == NIFTI_IMAGE_SHAPE


def test_nifti_image_from_stream(nifti_image_path: pathlib.Path) -> None:
    with open(nifti_image_path, "rb") as f:
        image = mioi.Image.from_stream(f)
    assert image.shape == NIFTI_IMAGE_SHAPE


def test_nifti_image_from_zipped_stream(zipped_nifti_path: pathlib.Path) -> None:
    with open(zipped_nifti_path, "rb") as f:
        image = mioi.Image.from_zipped_stream(
            f, gzipped=False, image_class=nib.Nifti1Image
        )
    assert image.shape == NIFTI_IMAGE_SHAPE


def test_nifti_gzipped_image_from_zipped_stream(
    zipped_nifti_gzipped_path: pathlib.Path,
) -> None:
    with open(zipped_nifti_gzipped_path, "rb") as f:
        image = mioi.Image.from_zipped_stream(
            f, gzipped=True, image_class=nib.Nifti1Image
        )
    assert image.shape == NIFTI_IMAGE_SHAPE


# flake8: noqa: E501
def test_numpy_ufuncs_on_dicom_image(dicom_image_dir: pathlib.Path) -> None:
    image = miod.DICOMImage.from_path(dicom_image_dir)
    assert image.shape == (TEST_IMAGE_SHAPE + (NUM_DUPLICATES,))
    image = image + 0.0
    image += 1.0
    image *= image
    assert isinstance(image, miod.DICOMImage)
    s = f"DICOMImage(shape: ({TEST_IMAGE_SHAPE[0]}, {TEST_IMAGE_SHAPE[1]}, {NUM_DUPLICATES}); spacing: (0.66, 0.66, {DIST}); dtype: float32)"
    assert str(image) == s
    mask = image == 0.0
    assert isinstance(mask, miod.DICOMImage)
    subimage = image[:20, :20, :]
    assert isinstance(subimage, miod.DICOMImage)
    image = image.astype(np.float16)
    assert isinstance(image, miod.DICOMImage)
    s = f"DICOMImage(shape: ({TEST_IMAGE_SHAPE[0]}, {TEST_IMAGE_SHAPE[1]}, {NUM_DUPLICATES}); spacing: (0.66, 0.66, {DIST}); dtype: float16)"
    assert str(image) == s


# flake8: noqa: E501
def test_numpy_ufuncs_on_image(image: mioi.Image) -> None:
    image += 1.0
    mask = image == 1.0
    assert isinstance(mask, mioi.Image)
    assert np.all(mask)
    image *= image
    assert np.all(image == 1.0)
    assert isinstance(image, mioi.Image)
    image = image[0:2]
    assert isinstance(image, mioi.Image)
    image[0:2] = 0.0
    s = f"Image(shape: {NIFTI_IMAGE_SHAPE}; spacing: (1, 1, 1); dtype: float32; orientation: RAS+)"
    assert str(image) == s
    _image: mioi.Image = np.squeeze(image).astype(np.float16)  # type: ignore[assignment]
    s = f"Image(shape: {NIFTI_IMAGE_SHAPE}; spacing: (1, 1, 1); dtype: float16; orientation: RAS+)"
    assert str(_image) == s
    image = image.ravel()
    assert isinstance(image, mioi.Image)
    image = image.view(mioi.Image)
    assert isinstance(image, mioi.Image)
    _image2 = image.view(np.ndarray)
    assert isinstance(image, np.ndarray)
    assert isinstance(_image2, np.ndarray)


def test_convert_to_torch(image: mioi.Image) -> None:
    torch = pytest.importorskip("torch")
    torch.as_tensor(image.torch_compatible())


def test_save_image(
    tmp_path_factory: pytest.TempPathFactory, image: mioi.Image
) -> None:
    save_path = tmp_path_factory.mktemp("save").resolve(strict=True)
    image.save(save_path / "test.nii.gz")


def test_save_image_npz(image: mioi.Image) -> None:
    with io.BytesIO() as buffer:
        image.to_npz(buffer)
        buffer.seek(0)
        new_image: mioi.Image = mioi.Image.from_npz(buffer)
        assert np.all(image == new_image)


def test_to_sitk(image: mioi.Image) -> None:
    assert isinstance(image.to_sitk(), sitk.Image)


def test_to_nibabel(image: mioi.Image) -> None:
    assert isinstance(image.to_nibabel(), nib.Nifti1Image)



"""Top-level package for pymedio."""

import logging

__title__ = "pymedio"
__description__ = "read arbitrary medical image formats in python"
__url__ = "https://github.com/jcreinhold/pymedio"
__author__ = """Jacob Reinhold"""
__email__ = "jcreinhold@gmail.com"
__version__ = "0.2.1"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Jacob Reinhold"

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""miscellaneous project functions
Author: Jacob Reinhold <jcreinhold@gmail.com>
"""

from __future__ import annotations

__all__ = [
    "check_uint_to_int",
    "ensure_4d",
    "flipxy_33",
    "flipxy_44",
    "get_metadata_from_ras_affine",
    "get_rotation_and_spacing_from_affine",
    "is_iterable",
    "to_f64",
    "unzip",
]

import builtins
import typing

import numpy as np
import numpy.typing as npt

import pymedio.typing as miot


# Matrices used to switch between LPS and RAS
def _create_flip_matrix(elements: typing.Tuple[builtins.float, ...]) -> npt.NDArray:
    matrix: npt.NDArray = to_f64(np.diag(elements))
    return matrix


def flipxy_33() -> npt.NDArray:
    return _create_flip_matrix((-1.0, -1.0, 1.0))


def flipxy_44() -> npt.NDArray:
    return _create_flip_matrix((-1.0, -1.0, 1.0, 1.0))


def is_iterable(x: typing.Any) -> builtins.bool:
    try:
        _ = iter(x)
    except TypeError:
        return False
    else:
        return True


def unzip(
    zipped_list: typing.Iterable[typing.Tuple[typing.Any, ...]]
) -> typing.Iterable[typing.Tuple[typing.Any, ...]]:
    return zip(*zipped_list)


def get_rotation_and_spacing_from_affine(
    affine: npt.NDArray,
) -> typing.Tuple[npt.NDArray, npt.NDArray]:
    # From https://github.com/nipy/nibabel/blob/master/nibabel/orientations.py
    rotation_zoom = affine[:3, :3]
    spacing = np.sqrt(np.sum(rotation_zoom * rotation_zoom, axis=0))
    rotation = rotation_zoom / spacing
    return rotation, spacing


def get_metadata_from_ras_affine(
    affine: npt.NDArray,
    *,
    is_2d: builtins.bool = False,
    lps: builtins.bool = True,
) -> typing.Tuple[miot.TripletFloat, miot.TripletFloat, miot.Direction]:
    direction_ras, spacing_array = get_rotation_and_spacing_from_affine(affine)
    origin_ras = affine[:3, 3]
    _flipxy_33 = flipxy_33()
    origin_lps = np.dot(_flipxy_33, origin_ras)
    direction_lps = np.dot(_flipxy_33, direction_ras)
    if is_2d:  # ignore orientation if 2D (1, W, H, 1)
        direction_lps = to_f64(np.diag((-1.0, -1.0)))
        direction_ras = to_f64(np.diag((1.0, 1.0)))
    origin_array = origin_lps if lps else origin_ras
    direction_array = direction_lps if lps else direction_ras
    direction_array = direction_array.flatten()
    # The following are to comply with typing hints
    # (there must be prettier ways to do this)
    ox, oy, oz = origin_array
    sx, sy, sz = spacing_array
    direction: miot.Direction
    if is_2d:
        d1, d2, d3, d4 = direction_array
        direction = d1, d2, d3, d4
    else:
        d1, d2, d3, d4, d5, d6, d7, d8, d9 = direction_array
        direction = d1, d2, d3, d4, d5, d6, d7, d8, d9
    origin = ox, oy, oz
    spacing = sx, sy, sz
    return origin, spacing, direction


def ensure_4d(
    array: npt.NDArray, *, num_spatial_dims: builtins.int | None = None
) -> npt.NDArray:
    """for PyTorch"""
    num_dimensions = array.ndim
    if num_dimensions == 4:
        pass
    elif num_dimensions == 5:  # hope (W, H, D, 1, C)
        if array.shape[-2] == 1:
            array = array[..., 0, :]
            array = array.transpose((3, 0, 1, 2))
        else:
            raise ValueError("5D is not supported for shape[-2] > 1")
    elif num_dimensions == 2:  # assume 2D monochannel (W, H)
        array = array[np.newaxis, ..., np.newaxis]  # (1, W, H, 1)
    elif num_dimensions == 3:  # 2D multichannel or 3D monochannel?
        if num_spatial_dims == 2:
            array = array[..., np.newaxis]  # (C, W, H, 1)
        elif num_spatial_dims == 3:  # (W, H, D)
            array = array[np.newaxis]  # (1, W, H, D)
        else:  # try to guess
            shape = array.shape
            maybe_rgb = 3 in (shape[0], shape[-1])
            if maybe_rgb:
                if shape[-1] == 3:  # (W, H, 3)
                    array = array.transpose((2, 0, 1))  # (3, W, H)
                array = array[..., np.newaxis]  # (3, W, H, 1)
            else:  # (W, H, D)
                array = array[np.newaxis]  # (1, W, H, D)
    else:
        message = f"{num_dimensions}D images not supported yet."
        raise NotImplementedError(message)
    if array.ndim != 4:
        raise RuntimeError("array couldn't be shaped to 4 dimensions")
    return array


def check_uint_to_int(array: npt.NDArray) -> npt.NDArray:
    """convert to int b/c PyTorch won't take uint16 nor uint32"""
    if array.dtype == np.uint16:
        return array.astype(np.int32)
    elif array.dtype == np.uint32:
        return array.astype(np.int64)
    else:
        return array


def to_f64(seq: npt.ArrayLike) -> npt.NDArray[np.float64]:
    """convert a sequence to a float64 array"""
    return np.asanyarray(seq, dtype=np.float64)



"""Functions for reading medical images
Taken from torchio and modified to output np.ndarray
https://github.com/fepegar/torchio
Author: Jacob Reinhold <jcreinhold@gmail.com>
"""

from __future__ import annotations

__all__ = [
    "read_affine",
    "read_image",
    "read_matrix",
    "write_image",
    "write_matrix",
]

import builtins
import gzip
import logging
import pathlib
import traceback
import typing
import warnings

import numpy as np
import numpy.typing as npt

try:
    import nibabel as nib
    import SimpleITK as sitk
except ImportError as imp_exn:
    imp_exn_msg = f"NiBabel and SimpleITK must be installed to use {__name__}."
    raise ImportError(imp_exn_msg) from imp_exn

import pymedio.typing as miot
import pymedio.utils as miou

# Image formats that are typically 2D
_2d_formats = [".jpg", ".jpeg", ".bmp", ".png", ".tif", ".tiff"]
IMAGE_2D_FORMATS = _2d_formats + [s.upper() for s in _2d_formats]

NibabelImageClass = typing.Type[
    typing.Union[
        nib.Nifti1Pair,
        nib.Nifti1Image,
        nib.Nifti2Pair,
        nib.Cifti2Image,
        nib.Nifti2Image,
        nib.Spm2AnalyzeImage,
        nib.Spm99AnalyzeImage,
        nib.AnalyzeImage,
        nib.Minc1Image,
        nib.Minc2Image,
        nib.MGHImage,
        nib.GiftiImage,
    ]
]

logger = logging.getLogger(__name__)


def read_image(
    path: miot.PathLike,
    *,
    dtype: npt.DTypeLike = np.float32,
    mmap: builtins.bool = False,
) -> miot.DataAffine:
    try:
        result = _read_sitk(path, dtype=dtype)
    except RuntimeError as exn1:  # try with NiBabel
        message = f"Error loading image with SimpleITK:\n{exn1}\n\nTrying NiBabel..."
        warnings.warn(message)
        try:
            result = _read_nibabel(path, dtype=dtype, mmap=mmap)
        except nib.loadsave.ImageFileError as exn2:
            message = (
                f"File '{path}' not understood."
                " Check supported formats by at"
                " https://simpleitk.readthedocs.io/en/master/IO.html#images"
                " and https://nipy.org/nibabel/api.html#file-formats"
            )
            raise RuntimeError(message) from exn2
    return result


def read_image_from_stream(
    stream: typing.BinaryIO,
    *,
    dtype: npt.DTypeLike = np.float32,
    gzipped: builtins.bool = False,
    image_class: typing.Optional[NibabelImageClass] = None,
) -> miot.DataAffine:
    """https://mail.python.org/pipermail/neuroimaging/2017-February/001345.html"""
    _stream = gzip.GzipFile(fileobj=stream) if gzipped else stream
    fh = nib.FileHolder(fileobj=_stream)
    if image_class is None:
        for cls in nib.imageclasses.all_image_classes:
            if hasattr(cls, "from_file_map"):
                try:
                    img = cls.from_file_map({"header": fh, "image": fh}, mmap=False)
                    break
                except Exception:
                    logger.debug(traceback.format_exc())
        else:
            raise RuntimeError("Couldn't open data stream.")
    else:
        img = image_class.from_file_map({"header": fh, "image": fh}, mmap=False)
    data = img.get_fdata(dtype=dtype)
    if data.ndim == 5:
        data = data[..., 0, :]
        data = data.transpose(3, 0, 1, 2)
    affine = img.affine
    return data, affine


def _read_nibabel(
    path: miot.PathLike,
    *,
    dtype: npt.DTypeLike = np.float32,
    mmap: builtins.bool = False,
) -> miot.DataAffine:
    img = nib.load(str(path), mmap=mmap)
    data = img.get_fdata(dtype=dtype)
    if data.ndim == 5:
        data = data[..., 0, :]
        data = data.transpose(3, 0, 1, 2)
    affine = miou.to_f64(img.affine)
    return data, affine


def _read_sitk(
    path: miot.PathLike, *, dtype: npt.DTypeLike = np.float32
) -> miot.DataAffine:
    if pathlib.Path(path).is_dir():  # assume DICOM
        image = _read_dicom_sitk(path)
    else:
        image = sitk.ReadImage(str(path))
    data, affine = sitk_to_array(image, dtype=dtype)
    return data, affine


def _read_dicom_sitk(directory: miot.PathLike) -> sitk.Image:
    directory = pathlib.Path(directory)
    if not directory.is_dir():  # unreachable if called from _read_sitk
        raise FileNotFoundError(f"Directory '{directory}' not found")
    reader = sitk.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(str(directory))
    if not dicom_names:
        message = f"The directory '{directory}' does not seem to contain DICOM files"
        raise FileNotFoundError(message)
    reader.SetFileNames(dicom_names)
    image = reader.Execute()
    return image


def read_shape(path: miot.PathLike) -> miot.Shape:
    reader = sitk.ImageFileReader()
    reader.SetFileName(str(path))
    reader.ReadImageInformation()
    num_channels = reader.GetNumberOfComponents()
    spatial_shape = reader.GetSize()
    num_dimensions = reader.GetDimension()
    if num_dimensions == 2:
        spatial_shape = *spatial_shape, 1
    elif num_dimensions == 4:  # assume bad NIfTI
        *spatial_shape, num_channels = spatial_shape
    sx, sy, sz = spatial_shape
    shape = (num_channels, sx, sy, sz)
    return shape


def read_affine(path: miot.PathLike) -> npt.NDArray:
    reader = get_reader(path)
    affine = get_ras_affine_from_sitk(reader)
    return affine


def get_reader(
    path: miot.PathLike, *, read: builtins.bool = True
) -> sitk.ImageFileReader:
    reader = sitk.ImageFileReader()
    reader.SetFileName(str(path))
    if read:
        reader.ReadImageInformation()
    return reader


def write_image(
    array: npt.NDArray,
    affine: npt.NDArray,
    path: miot.PathLike,
    *,
    squeeze: builtins.bool = True,
    **write_sitk_kwargs: builtins.bool,
) -> None:
    if squeeze:
        array = array.squeeze()
    try:
        _write_sitk(array, affine, path, **write_sitk_kwargs)
    except RuntimeError:  # try with NiBabel
        _write_nibabel(array, affine, path)


def _write_nibabel(
    array: npt.NDArray,
    affine: npt.NDArray,
    path: miot.PathLike,
) -> None:
    """
    Expects a path with an extension that can be used by nibabel.save
    to write a NIfTI-1 image, such as '.nii.gz' or '.img'
    """
    num_components = array.shape[0]
    # NIfTI components must be at the end, in a 5D array
    if num_components == 1:
        array = array[0]
    elif array.ndim == 4:
        array = array.transpose((1, 2, 3, 0))
    suffix = pathlib.Path(str(path).replace(".gz", "")).suffix
    if ".nii" in suffix:
        img = nib.Nifti1Image(np.asanyarray(array), affine)
    elif ".hdr" in suffix or ".img" in suffix:
        img = nib.Nifti1Pair(np.asanyarray(array), affine)
    else:
        raise nib.loadsave.ImageFileError
    if num_components > 1:
        img.header.set_intent("vector")
    img.header["qform_code"] = 1
    img.header["sform_code"] = 0
    nib.save(img, str(path))


def _write_sitk(
    array: npt.NDArray,
    affine: npt.NDArray,
    path: miot.PathLike,
    *,
    use_compression: builtins.bool = True,
    is_multichannel: builtins.bool = False,
) -> None:
    path = pathlib.Path(path)
    if path.suffix in (".png", ".jpg", ".jpeg", ".bmp"):
        msg = f"Casting to uint8 before saving to {path}"
        warnings.warn(msg, RuntimeWarning)
        array = array.astype(dtype=np.uint8, copy=False)
    image = array_to_sitk(array, affine, is_multichannel=is_multichannel)
    sitk.WriteImage(image, str(path), use_compression)


def read_matrix(path: miot.PathLike) -> npt.NDArray:
    """Read an affine transform and return array"""
    path = pathlib.Path(path)
    suffix = path.suffix
    if suffix in (".tfm", ".h5"):  # ITK
        array = _read_itk_matrix(path)
    elif suffix in (".txt", ".trsf"):  # NiftyReg, blockmatching
        array = _read_niftyreg_matrix(path)
    else:
        raise ValueError(f"Unknown suffix for transform file: '{suffix}'")
    return array


def write_matrix(matrix: npt.NDArray, path: miot.PathLike) -> None:
    """Write an affine transform."""
    path = pathlib.Path(path)
    suffix = path.suffix
    if suffix in (".tfm", ".h5"):  # ITK
        _write_itk_matrix(matrix, path)
    elif suffix in (".txt", ".trsf"):  # NiftyReg, blockmatching
        _write_niftyreg_matrix(matrix, path)


def _to_itk_convention(matrix: npt.NDArray) -> npt.NDArray:
    """RAS to LPS"""
    _flipxy_44 = miou.flipxy_44()
    matrix = np.dot(_flipxy_44, matrix)
    matrix = np.dot(matrix, _flipxy_44)
    matrix = np.linalg.inv(matrix)
    return matrix


def _from_itk_convention(matrix: npt.NDArray) -> npt.NDArray:
    """LPS to RAS"""
    _flipxy_44 = miou.flipxy_44()
    matrix = np.dot(matrix, _flipxy_44)
    matrix = np.dot(_flipxy_44, matrix)
    matrix = np.linalg.inv(matrix)
    return matrix


def _read_itk_matrix(path: miot.PathLike) -> npt.NDArray:
    """Read an affine transform in ITK's .tfm format"""
    transform = sitk.ReadTransform(str(path))
    parameters = transform.GetParameters()
    rotation_params = parameters[:9]
    rotation_matrix = miou.to_f64(rotation_params).reshape(3, 3)
    translation_params = parameters[9:]
    translation_vector = miou.to_f64(translation_params).reshape(3, 1)
    matrix: np.ndarray = np.hstack([rotation_matrix, translation_vector])
    homogeneous_matrix_lps: np.ndarray = np.vstack([matrix, [0.0, 0.0, 0.0, 1.0]])
    homogeneous_matrix_ras = _from_itk_convention(homogeneous_matrix_lps)
    return homogeneous_matrix_ras


def _write_itk_matrix(matrix: npt.NDArray, tfm_path: miot.PathLike) -> None:
    """The tfm file contains the matrix from floating to reference."""
    transform = _matrix_to_itk_transform(matrix)
    transform.WriteTransform(str(tfm_path))


def _matrix_to_itk_transform(
    matrix: npt.NDArray, *, dims: builtins.int = 3
) -> sitk.AffineTransform:
    matrix = _to_itk_convention(matrix)
    rotation = matrix[:dims, :dims].ravel().tolist()
    translation = matrix[:dims, 3].tolist()
    transform = sitk.AffineTransform(rotation, translation)
    return transform


def _read_niftyreg_matrix(trsf_path: miot.PathLike) -> npt.NDArray:
    """Read a NiftyReg matrix and return it as a torch.Tensor"""
    matrix: np.ndarray = np.loadtxt(trsf_path)
    matrix = np.linalg.inv(matrix)
    return matrix


def _write_niftyreg_matrix(matrix: npt.NDArray, txt_path: miot.PathLike) -> None:
    """Write an affine transform in NiftyReg's .txt format (ref -> flo)"""
    matrix = np.linalg.inv(matrix)
    np.savetxt(txt_path, matrix, fmt="%.8f")


def array_to_sitk(
    array: npt.NDArray,
    affine: npt.NDArray,
    *,
    is_multichannel: builtins.bool = False,
) -> sitk.Image:
    """Create a SimpleITK image from an array and a 4x4 affine matrix."""
    ndim = array.ndim
    array = np.asanyarray(array)
    affine = miou.to_f64(affine)
    image = sitk.GetImageFromArray(array.transpose(), isVector=is_multichannel)
    is_2d = (ndim == 3 and is_multichannel) or (ndim == 2 and not is_multichannel)
    origin, spacing, direction = miou.get_metadata_from_ras_affine(
        affine,
        is_2d=is_2d,
    )
    image.SetOrigin(origin)
    image.SetSpacing(spacing)
    image.SetDirection(direction)
    num_spatial_dims = 2 if is_2d else 3
    offset = 1 if is_multichannel else 0
    if is_multichannel:
        num_components = array.shape[0]
        if (_n_comp := image.GetNumberOfComponentsPerPixel()) != num_components:
            msg = f"sitk components {_n_comp} != array components {num_components}"
            raise RuntimeError(msg)
    spatial_dims = array.shape[offset : offset + num_spatial_dims]
    if image.GetSize() != spatial_dims:
        raise RuntimeError(f"{image.GetSize()} != {spatial_dims}")
    return image


def sitk_to_array(
    image: sitk.Image, *, dtype: npt.DTypeLike = np.float32
) -> miot.DataAffine:
    array_view = sitk.GetArrayViewFromImage(image)
    data: np.ndarray = np.asarray(array_view, dtype=dtype).transpose()
    num_components = image.GetNumberOfComponentsPerPixel()
    input_spatial_dims = image.GetDimension()
    if input_spatial_dims == 5:  # probably a bad NIfTI (1, sx, sy, sz, c)
        # Try to fix it
        num_components = data.shape[-1]
        data = data[0]
        data = data.transpose(3, 0, 1, 2)
    if num_components > 1 and data.shape[0] != num_components:
        raise RuntimeError(f"{data.shape[0]} != {num_components}")
    affine = get_ras_affine_from_sitk(image)
    return data, affine


def get_ras_affine_from_sitk(
    sitk_object: sitk.Image | sitk.ImageFileReader, *, dtype: npt.DTypeLike = np.float64
) -> npt.NDArray:
    spacing: np.ndarray = np.asarray(sitk_object.GetSpacing(), dtype=dtype)
    direction_lps: np.ndarray = np.asarray(sitk_object.GetDirection(), dtype=dtype)
    origin_lps: np.ndarray = np.asarray(sitk_object.GetOrigin(), dtype=dtype)
    direction_length = len(direction_lps)
    if direction_length == 9:
        rotation_lps = direction_lps.reshape(3, 3)
    elif direction_length == 4:  # ignore last dimension if 2D (1, W, H, 1)
        rotation_lps_2d = direction_lps.reshape(2, 2)
        rotation_lps = np.eye(3, dtype=dtype)
        rotation_lps[:2, :2] = rotation_lps_2d
        spacing = np.append(spacing, 1)
        origin_lps = np.append(origin_lps, 0)
    elif direction_length == 16:  # probably a bad NIfTI. Let's try to fix it
        rotation_lps = direction_lps.reshape(4, 4)[:3, :3]
        spacing = spacing[:-1]
        origin_lps = origin_lps[:-1]
    else:
        raise RuntimeError(f"Invalid direction length: {direction_length}")
    _flipxy_33 = miou.flipxy_33()
    rotation_ras = np.dot(_flipxy_33, rotation_lps)
    rotation_ras_zoom = rotation_ras * spacing
    translation_ras = np.dot(_flipxy_33, origin_lps)
    affine: np.ndarray = np.eye(4, dtype=dtype)
    affine[:3, :3] = rotation_ras_zoom
    affine[:3, 3] = translation_ras
    return affine



"""Base image class for image containers
Author: Jacob Reinhold <jcreinhold@gmail.com>
"""

from __future__ import annotations

__all__ = ["BasicImage"]

import builtins
import numbers
import typing
import warnings

import numpy as np
import numpy.typing as npt

import pymedio.typing as miot
import pymedio.utils as miou

_UfuncMethod = typing.Literal[
    "__call__", "reduce", "reduceat", "accumulate", "outer", "inner"
]
_Image = typing.TypeVar("_Image", bound="BasicImage")


class BasicImage(np.ndarray):

    _HANDLED_TYPES = (np.ndarray, numbers.Number)
    _affine: npt.NDArray

    def __new__(
        cls: typing.Type[_Image],
        data: npt.ArrayLike,
        affine: npt.NDArray | None = None,
    ) -> _Image:
        obj = cls._check_data(data).view(cls)
        obj._affine = cls._check_affine(affine)
        obj._affine.flags.writeable = False
        return obj

    def __array_finalize__(self, obj: builtins.object) -> None:
        if obj is None:
            return
        self._affine = self._check_affine(getattr(obj, "_affine", None))
        self._affine.flags.writeable = False

    @property
    def str_properties(self) -> typing.List[builtins.str]:
        return [
            f"shape: {self.shape}",
            f"spacing: {self.get_spacing_string()}",
            f"dtype: {self.dtype.name}",
        ]

    def __str__(self) -> builtins.str:
        properties = "; ".join(self.str_properties)
        string = f"{self.__class__.__name__}({properties})"
        return string

    def __array_ufunc__(
        self,
        ufunc: np.ufunc,
        method: _UfuncMethod,
        *inputs: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.Any:
        out = kwargs.get("out", ())
        for x in inputs + out:
            if not isinstance(x, self._HANDLED_TYPES + (self.__class__,)):
                return NotImplemented
        affine = self.affine
        is_cls = lambda y: isinstance(y, self.__class__)  # noqa: E731
        ufunc_args = tuple(x.view(np.ndarray) if is_cls(x) else x for x in inputs)
        if out:
            kwargs["out"] = tuple(x.view(np.ndarray) if is_cls(x) else x for x in out)
            if len(out) == 1 and is_cls(out[0]):
                affine = out[0].affine
        result = getattr(ufunc, method)(*ufunc_args, **kwargs)
        if isinstance(result, tuple):
            return tuple(self.__class__(x, affine) for x in result)
        elif method == "at":
            return None
        elif isinstance(result, np.ndarray):
            return typing.cast(_Image, self.__class__(result, affine))
        else:
            return result

    @property
    def affine(self) -> npt.NDArray:
        return self._affine

    @affine.setter
    def affine(self, new_affine: npt.NDArray) -> None:
        self._affine = self._check_affine(new_affine)

    @property
    def direction(self) -> miot.Direction:
        _, _, direction = miou.get_metadata_from_ras_affine(self.affine, lps=False)
        return direction

    @property
    def spacing(self) -> typing.Tuple[builtins.float, ...]:
        """Voxel spacing in mm."""
        _, spacing = miou.get_rotation_and_spacing_from_affine(self.affine)
        return tuple(spacing)

    @property
    def origin(self) -> typing.Tuple[builtins.float, ...]:
        """Center of first voxel in array, in mm."""
        return tuple(self.affine[:3, 3])

    @property
    def memory(self) -> builtins.int:
        """Number of Bytes that the tensor takes in the RAM."""
        mem: builtins.int = np.prod(self.shape) * self.itemsize
        return mem

    def get_spacing_string(self) -> builtins.str:
        strings = [f"{n:.2g}" for n in self.spacing]
        string = f'({", ".join(strings)})'
        return string

    @staticmethod
    def _check_data(data: npt.ArrayLike) -> npt.NDArray:
        data = np.asanyarray(data)
        if np.isnan(data).any():
            warnings.warn("NaNs found in data", RuntimeWarning)
        if any(d == 0 for d in data.shape):
            msg = f"Data must have all non-zero dimensions. Got {data.shape}."
            raise ValueError(msg)
        return data

    @staticmethod
    def _check_affine(affine: npt.NDArray | None) -> npt.NDArray:
        if affine is None:
            return np.eye(4, dtype=np.float64)
        if not isinstance(affine, np.ndarray):
            bad_type = type(affine)
            raise TypeError(f"Affine must be a NumPy array, not {bad_type}")
        if affine.shape != (4, 4):
            bad_shape = affine.shape
            raise ValueError(f"Affine shape must be (4, 4), not {bad_shape}")
        return miou.to_f64(affine)

    def to_npz(self, file: miot.PathLike | typing.BinaryIO) -> None:
        np.savez_compressed(file, data=self.base, affine=self.affine)

    @classmethod
    def from_npz(
        cls, file: miot.PathLike | typing.BinaryIO, **np_load_kwargs
    ) -> _Image:
        _data = np.load(file, **np_load_kwargs)
        return cls(_data["data"], affine=_data["affine"])

    def torch_compatible(self) -> npt.NDArray:
        return miou.ensure_4d(miou.check_uint_to_int(self.view(np.ndarray)))
